;; Generated package description from jsonrpc.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "jsonrpc" "1.0.29.0.20260623.0" "JSON-RPC library" '((emacs "25.2")) :commit "07c786f8618b2847f7951ed0de3ccb45377e2bf8" :url "https://elpa.gnu.org/nongnu/jsonrpc.html" :authors '(("João Távora" . "joaotavora@gmail.com")) :maintainer '("João Távora" . "joaotavora@gmail.com") :keywords '("processes" "languages" "extensions"))
